﻿using Nemerle.Collections;
using Nemerle.Text;
using Nemerle.Utility;

using System;
using System.Collections.Generic;
using System.Console;
using System.Linq;

public variant Token
{
  | Number   { value : int; }
  | Operator { name  : string; }
  | OpenBrace
  | CloseBrace
  | Error

  public override ToString() : string
  {
    match (this)
    {
      | Number(value)  => $"$value"
      | Operator(name) => $"$name"
      | OpenBrace      => "("
      | CloseBrace     => ")"
      | Error          => "Error"
    }
  }

  public this(startPos : int)
  {
    StartPos = startPos;
  }
  
  public StartPos : int;
}

public class CalcParser
{
  public mutable Error : int * string;
  
  public Lexer(text : string) : list[Token]
  {
    mutable index = 0;
    
    def peek() : char
    {
      if (index >= text.Length) '\0' else text[index]
    }

    def read() : char
    {
      def ch = peek();
      when (index < text.Length)
        index++;
      ch
    }

    def isDigit(ch) { ch >= '0' && ch <= '9' }

    def loop(res : list[Token]) : list[Token]
    {
      def number(ch : char, accumulator : int = 0) : int
      {
        def highOrderValue    = accumulator * 10;
        def currentOrderValue = ch - '0' : int;
        def currentValue      = highOrderValue + currentOrderValue;
    
        if (isDigit(peek()))
          number(read(), currentValue);
        else
          currentValue
      }

      def error(startPos)
      {
        Error = (startPos, "ожидается число или оператор");
        (Token.Error(startPos) :: res).Reverse()
      }
      
      def startPos = index;
      def ch       = read();

      match (ch)
      {
        | '('                   => loop(Token.OpenBrace (startPos) :: res)
        | ')'                   => loop(Token.CloseBrace(startPos) :: res)
        | ' ' | '\t'            => loop(res) // игнорируем пробелы
        | '+' | '-' | '*' | '/' => loop(Token.Operator(startPos, ch.ToString()) :: res)
        | '\0'                  => res.Reverse()
        | _ when isDigit(ch)    => loop(Token.Number(startPos, number(ch)) :: res)
        | _                     => error(startPos)
      }
    }
    
    loop([])
  }
  
  public ParseAndEval(tokens : list[Token]) : option[int]
  {
    def loop2(input : option[int] * list[Token], parseSubExpr, op1, f1, op2, f2)
    {
      match (input)
      {
        | (Some(first), Token.Operator(op) as tok :: tokens) when op == op1 with f = f1
        | (Some(first), Token.Operator(op) as tok :: tokens) when op == op2 with f = f2 =>
          match (parseSubExpr(tokens))
          {
            | (Some(second), tokens) => 
              loop2((f(first, second, tok.StartPos), tokens), parseSubExpr, op1, f1, op2, f2)
              
            | (None, _)              => (None(), tokens)
          }

        | (Some as firstOpt, tokens) => (firstOpt, tokens)
        | (None, tokens)             => (None(), tokens)
      }
    }
    def loop(input : option[int] * list[Token], parseSubExpr, op1, f1, op2, f2)
    {
      loop2(input, parseSubExpr, op1, (a, b, _) => Some(f1(a, b)), 
                                 op2, (a, b, _) => Some(f2(a, b)))
    }
    def parseSimpleExpr(tokens : list[Token]) : option[int] * list[Token]
    {
      | Token.Operator("-") :: tail => 
        match (parseSimpleExpr(tail))
        {
          | (Some(value), tokens) => (Some(-value), tokens)
          | failResult            => failResult
        }
        
      | Token.Number(value) :: tail => (Some(value), tail)
      | Token.OpenBrace as openBraceTok :: tail => 
        match (parseSumOrSub(tail))
        {
          | (Some(value), Token.CloseBrace :: tail) => (Some(value), tail)
          | (Some, _) => // за подвыражением не идет закрывающей скобки!
            Error = (openBraceTok.StartPos, "Не закрытая скобка");
            (None(), tokens)
            
          | failResult => failResult
        }
      
      | Token.Error         :: _    => (None(), tokens)
      | unexpectedToken :: _        =>
        Error = (unexpectedToken.StartPos, "Ожидается число");
        (None(), tokens)
      
      | _                           =>
        Error = (-1, "В конце выяражения ожидается число");
        (None(), tokens)      
    }
    and parseMulOrDiv(tokens : list[Token]) : option[int] * list[Token]
    {
      def div(a, b, pos)
      {
        if (b == 0)
        {
          Error = (pos, "деление на 0");
          None()
        }
        else Some(a / b)
      }
      loop2(parseSimpleExpr(tokens), parseSimpleExpr, "*", (a, b, _) => Some(a * b), "/", div)
    }
    and parseSumOrSub(tokens : list[Token]) : option[int] * list[Token]
    {
      loop(parseMulOrDiv(tokens), parseMulOrDiv, "+", _ + _, "-", _ - _)
    }
    
    match (parseSumOrSub(tokens))
    {
      | (None, _)                     => None()
      | (Some as value, [])           => value
      | (Some, unexpectedToken :: _)  => 
        Error = (unexpectedToken.StartPos, "Нераспозннаый токен");
        None()
    }
  }

  public ParseAndEval(text : string) : option[int]
  {
    ParseAndEval(Lexer(text))
  }
}

module Program
{
  Main() : void
  {
    def readInput()
    {
      def inputString = ReadLine();
      
      unless (inputString == "")
      {
        def calc = CalcParser();
        
        match (calc.ParseAndEval(inputString))
        {
          | Some(value) => WriteLine(value);
          | None        => 
            def (pos, message) = calc.Error;
            WriteLine(string(' ', if (pos < 0) inputString.Length - 1 else pos) + "^");
            WriteLine(message);
        }
          
        readInput();
      }
    }

    readInput();
  }
}