﻿using System;
using System.Console;
using Nemerle.Utility;

[Record]
variant Token
{
  | Number   { value : int; }
  | Operator { name  : string; }
  | OpenBrace
  | CloseBrace
  | Eof
  | Error
  
  public StartPos : int;
  
  #region object  Members
  
  public override ToString() : string
  {
    match (this)
    {
      | Number(value)  => $"$value"
      | Operator(name) => $"$name"
      | OpenBrace => "("
      | CloseBrace => ")"
      | Eof => "."
      | Error => "Error"
    }
  }
  
  #endregion object  Members
  
}

module Program
{
  Main() : void
  {
    def calc(text : string)
    {
      def lexer(text : string) : list[Token]
      {
        mutable index = 0;
        
        def peek()
        {
          if (index >= text.Length) '\0' else text[index];
        }
        def read()
        {
          def ch = peek();
          when (index < text.Length)
            index++;
          ch
        }
        def isNumber(ch) { ch >= '0' && ch <= '9' }
        def number(ch : char, num = 0) : int
        {
          def num = num * 10 + (ch - '0' : int);
          
          if (isNumber(peek()))
            number(read(), num);
          else
            num
        }
        
        def loop(res : list[Token]) : list[Token]
        {
          def ch = read();
          match (ch)
          {
            | '('                   => loop(Token.OpenBrace(index)  :: res)
            | ')'                   => loop(Token.CloseBrace(index) :: res)
            | ' ' | '\t'            => loop(res) // игнорируем пробелы
            | '+' | '-' | '*' | '/' => loop(Token.Operator(index, ch.ToString()) :: res)
            | '\0'                  => res.Reverse()
            | _ when isNumber(ch)   => loop(Token.Number(index, number(ch)) :: res)
            | _ =>
              WriteLine(string(' ', index - 1) + "^");
              WriteLine($"ожидается число или оператр (в позиции $index)");
              (Token.Error(index) :: res).Reverse()
          }
        }
        
        loop([])
      }
      
      def parseAndEval(tokens : list[Token])
      {
        def eof = Token.Eof(if (tokens.IsEmpty) 1 else tokens.Last.StartPos);
        mutable current = tokens;
        
        def peek()
        {
          match (current)
          {
            | null      => Token.Error(1)
            | head :: _ => head
            | []        => eof
          }
        }
        def read()
        {
          match (current)
          {
            | null      => Token.Error(1)
            | head :: tail => 
              current = tail;
              head
              
            | []        => eof
          }
        }
        def error(token : Token, message : string) : int
        {
          WriteLine(string(' ', token.StartPos - 1) + "^");
          WriteLine(message);
          current = null;
          0
        }
        // SimpleExpr = Number / '(' SumOrSub ')'
        // MulOrDive  = SimpleExpr (('*' / '/') SimpleExpr)*
        // SumOrSub   = MulOrDive  (('*' / '/') MulOrDive)*
        
        def parseSimpleExpr() : int
        {
          def tok = read();
          match (tok)
          {
            | Error => 0
            | Operator("-") => - parseSimpleExpr()
            | Number(val)   => val
            | OpenBrace     => 
              def value = parseSumOrSub();
              def toc = read();
              if (toc is Token.CloseBrace || toc is Token.Error)
                value
              else
                error(tok, "Не закрытая скобка!")
            
            | Eof          => error(tok, "В конце выражения должно стоять число")
            | CloseBrace   => error(tok, "Не ожидаемая ')'")
            | Operator(op) => error(tok, $"Не ожидаемый оператор '$op'")
          }
        }        
        and parseMulOrDive() : int
        {
          def loop(firstValue)
          {
            def tok = peek();
            
            match (tok)
            {
              | Operator("*") => 
                _ = read();
                def secondValue = parseSimpleExpr();
                loop(firstValue * secondValue)

              | Operator("/") => 
                _ = read();
                def secondValue = parseSimpleExpr();
                if (secondValue == 0)
                  error(tok, "Делитель не должен быть нулем!")
                else
                  loop(firstValue / secondValue)
                  
                
              | Error | Eof | Operator | CloseBrace => firstValue
              | OpenBrace   => error(tok, "Не ожидаемая '('")
              | Number(val) => error(tok, $"Значению '$val' не предшествует один из операторов")
            }
          }
          
          def firstValue = parseSimpleExpr();
          loop(firstValue)
        }
        and parseSumOrSub() : int
        {
          def loop(firstValue)
          {
            def tok = peek();

            match (tok)
            {
              | Operator("+") => 
                _ = read();
                def secondValue = parseMulOrDive();
                loop(firstValue + secondValue)

              | Operator("-") => 
                _ = read();
                def secondValue = parseMulOrDive();
                loop(firstValue - secondValue)
                
              | Error | Eof | CloseBrace => firstValue
              | OpenBrace    => error(tok, "Не ожидаемая '('")
              | Operator(op) => error(tok, $"Не ожидаемый оператор '$op'")
              | Number(val)  => error(tok, $"Значению '$val' не предшествует один из операторов")
            }
          }
          
          def firstValue = parseMulOrDive();
          loop(firstValue)
        }
        
        def parse() : int
        {
          def value = parseSumOrSub();
          def tok = peek();
          
          when (!(tok is Token.Eof) && !(tok is Token.Error))
            _ = error(peek(), $"Не ожидаемый токен '$(peek())'");
            
          value
        }
        
        def res = parse();
        when (current != null)
        {
          WriteLine("Результат:");
          WriteLine(res);
        }
      }
      
      def res = lexer(text);
      //WriteLine("Lexems:");
      //WriteLine(res);
      unless (res.IsEmpty || res.Last is Token.Error)
        parseAndEval(res);
    }
    
    def readInput()
    {
      def res = ReadLine();
      
      unless (res == "")
      {
        //WriteLine("Вы ввели:");
        //WriteLine(res);
        calc(res);
        readInput();
      }
    }
    
    readInput();
  }
}

        /*def getOperatorFun(operator : string) : int * int -> int
        {
          | "+" => def operatorFunk(x, y) { x + y } operatorFunk
          | "-" => def operatorFunk(x, y) { x - y } operatorFunk
          | "*" => def operatorFunk(x, y) { x * y } operatorFunk
          | "/" => def operatorFunk(x, y) { x / y } operatorFunk
          | _   => 
            WriteLine("Оператор '{0}' не поддерживается!", operator);
            def fake(_, _) { 0 }
            fake
        }*/